@@ -1,11 +1,11 @@
 /************************************************************
- * HMMER - Biological sequence analysis with profile-HMMs
+ * HMMER - Biological sequence analysis with profile HMMs
  * Copyright (C) 1992-1998 Washington University School of Medicine
- *
- *   This source code is distributed under the terms of the
- *   GNU General Public License. See the files COPYING and
- *   GNULICENSE for details.
- *
+ * All Rights Reserved
+ * 
+ *     This source code is distributed under the terms of the
+ *     GNU General Public License. See the files COPYING and LICENSE
+ *     for details.
  ************************************************************/
 
 /* hmmio.c
@@ -160,6 +160,8 @@ HMMFileOpen(char *hmmfile, char *env)
   HMMFILE     *hmmfp;
   unsigned int magic;
   char         buf[512];
+  char        *gsifile;
+  char        *dir;        /* dir name in which HMM file was found */
 
   hmmfp = (HMMFILE *) MallocOrDie (sizeof(HMMFILE));
   hmmfp->f          = NULL; 
@@ -170,12 +172,33 @@ HMMFileOpen(char *hmmfile, char *env)
   /* Open the file. Look in current directory.
    * If that doesn't work, check environment var for
    * a second possible directory (usually the location
-   * of a system-wide HMM library)
+   * of a system-wide HMM library).
+   * Using dir name if necessary, construct correct GSI file name.
    */
-  if ((hmmfp->f = fopen(hmmfile, "r"))       == NULL &&
-      (hmmfp->f = EnvFileOpen(hmmfile, env)) == NULL)
-    return NULL;
+  hmmfp->f   = NULL;
+  hmmfp->gsi = NULL;
+  if ((hmmfp->f = fopen(hmmfile, "r")) != NULL)
+    {
+      gsifile = MallocOrDie(sizeof(char) * (strlen(hmmfile) + 5));
+      sprintf(gsifile, "%s.gsi", hmmfile);
+    }
+  else if ((hmmfp->f = EnvFileOpen(hmmfile, env, &dir)) != NULL)
+    {
+      char *full;
+      full    = FileConcat(dir, hmmfile);
+      gsifile = MallocOrDie(sizeof(char) * (strlen(full) + strlen(hmmfile) + 5));
+      sprintf(gsifile, "%s.gsi", full);
+      free(full);
+    }
+  else return NULL;
   
+  /* Open the GSI index file. If it doesn't exist,
+   * hmmfp->gsi stays NULL.
+   */
+  SQD_DPRINTF1(("Opening gsifile %s...\n", gsifile));
+  hmmfp->gsi = GSIOpen(gsifile);
+  free(gsifile);
+
   /* Check for binary or byteswapped binary format
    * by peeking at first 4 bytes.
    */ 
@@ -191,6 +214,7 @@ HMMFileOpen(char *hmmfile, char *env)
     return hmmfp;
   } 
   else if (magic == v20swap) { 
+    SQD_DPRINTF1(("Opened a HMMER 2.0 binary file [byteswapped]\n"));
     hmmfp->parser    = read_bin20hmm;
     hmmfp->is_binary = TRUE;
     hmmfp->byteswap  = TRUE;
@@ -291,7 +315,8 @@ HMMFileRead(HMMFILE *hmmfp, struct plan7_s **ret_hmm)
 void
 HMMFileClose(HMMFILE *hmmfp)
 {
-  if (hmmfp->f   != NULL) fclose(hmmfp->f);
+  if (hmmfp->f   != NULL) fclose(hmmfp->f);      
+  if (hmmfp->gsi != NULL) GSIClose(hmmfp->gsi);
   free(hmmfp);
 }
 void 
@@ -299,7 +324,38 @@ HMMFileRewind(HMMFILE *hmmfp)
 {
   rewind(hmmfp->f);
 }
-			       
+void
+HMMFilePositionByOffset(HMMFILE *hmmfp, long offset)
+{
+  if (fseek(hmmfp->f, offset, SEEK_SET) != 0) PANIC;
+}
+int
+HMMFilePositionByName(HMMFILE *hmmfp, char *name)
+{				/* limitation: name < 32 char  */
+  char       hmmfile[32];	/* ignored.                    */
+  long       offset;		/* offset in hmmfile, from GSI */
+  int        fmt;		/* ignored.                    */
+
+  if (hmmfp->gsi == NULL) return 0; /* can only position if we have GSI index   */
+  if (strlen(name) >= 32) return 0; /* GSI files can only handle keys < 32 char */
+  if (! GSIGetOffset(hmmfp->gsi, name, hmmfile, &fmt, &offset)) return 0; 
+  HMMFilePositionByOffset(hmmfp, offset);
+  return 1;
+}
+int 
+HMMFilePositionByIndex(HMMFILE *hmmfp, int idx)
+{				/* idx runs from 0..nhmm-1 */
+  char       hmmname[32];	/* ignored */
+  sqd_uint32 offset;
+  sqd_uint16 filenum;
+
+  if (idx >= hmmfp->gsi->recnum) return 0; /* idx out of bounds */
+  if (idx <  0)                  return 0; /* idx out of bounds */
+  if (fseek(hmmfp->gsi->gsifp, (1 + hmmfp->gsi->nfiles + idx) * GSI_RECSIZE, SEEK_SET) != 0) PANIC;
+  GSIGetRecord(hmmfp->gsi, hmmname, &filenum, &offset);
+  HMMFilePositionByOffset(hmmfp, (long) offset);
+  return 1;
+}
 
 /*****************************************************************
  * HMM output API:
@@ -327,18 +383,26 @@ WriteAscHMM(FILE *fp, struct plan7_s *hmm)
   /* write header information
    */
   fprintf(fp, "NAME  %s\n", hmm->name);
-  fprintf(fp, "DESC  %s\n",
-	  (hmm->flags & PLAN7_DESC) ? hmm->desc : ""); 
+  if (hmm->flags & PLAN7_ACC)
+    fprintf(fp, "ACC   %s\n", hmm->acc);
+  if (hmm->flags & PLAN7_DESC) 
+    fprintf(fp, "DESC  %s\n", hmm->desc);
   fprintf(fp, "LENG  %d\n", hmm->M);
   fprintf(fp, "ALPH  %s\n",   
 	  (Alphabet_type == hmmAMINO) ? "Amino":"Nucleic");
-  fprintf(fp, "RF    %s\n",
-          (hmm->flags & PLAN7_RF) ? "yes" : "no");
-  fprintf(fp, "CS    %s\n",
-          (hmm->flags & PLAN7_CS) ? "yes" : "no");
+  fprintf(fp, "RF    %s\n", (hmm->flags & PLAN7_RF)  ? "yes" : "no");
+  fprintf(fp, "CS    %s\n", (hmm->flags & PLAN7_CS)  ? "yes" : "no");
+  fprintf(fp, "MAP   %s\n", (hmm->flags & PLAN7_MAP) ? "yes" : "no");
   multiline(fp, "COM   ", hmm->comlog);
   fprintf(fp, "NSEQ  %d\n", hmm->nseq);
   fprintf(fp, "DATE  %s\n", hmm->ctime); 
+  fprintf(fp, "CKSUM %d\n", hmm->checksum);
+  if (hmm->flags & PLAN7_GA)
+    fprintf(fp, "GA    %.1f %.1f\n", hmm->ga1, hmm->ga2);
+  if (hmm->flags & PLAN7_TC)
+    fprintf(fp, "TC    %.1f %.1f\n", hmm->tc1, hmm->tc2);
+  if (hmm->flags & PLAN7_NC)
+    fprintf(fp, "NC    %.1f %.1f\n", hmm->nc1, hmm->nc2);
 
   /* Specials
    */
@@ -379,12 +443,12 @@ WriteAscHMM(FILE *fp, struct plan7_s *hmm)
   fprintf(fp, "%6s\n", prob2ascii(hmm->tbd1, 1.0));
   for (k = 1; k <= hmm->M; k++)
     {
-				/* Line 1: k and match emissions */
+				/* Line 1: k, match emissions, map */
       fprintf(fp, " %5d ", k);
       for (x = 0; x < Alphabet_size; x++) 
         fprintf(fp, "%6s ", prob2ascii(hmm->mat[k][x], hmm->null[x]));
+      if (hmm->flags & PLAN7_MAP) fprintf(fp, "%5d", hmm->map[k]);
       fputs("\n", fp);
-
 				/* Line 2: RF and insert emissions */
       fprintf(fp, " %5c ", hmm->flags & PLAN7_RF ? hmm->rf[k] : '-');
       for (x = 0; x < Alphabet_size; x++) 
@@ -412,20 +476,35 @@ WriteBinHMM(FILE *fp, struct plan7_s *hmm)
   int k;
 
   /* ye olde magic number */
-  fwrite((char *) &(v20magic), sizeof(long), 1, fp);
+  fwrite((char *) &(v20magic), sizeof(unsigned int), 1, fp);
 
   /* header section
    */
   fwrite((char *) &(hmm->flags),    sizeof(int),  1,   fp);
   write_bin_string(fp, hmm->name);
+  if (hmm->flags & PLAN7_ACC)  write_bin_string(fp, hmm->acc);
   if (hmm->flags & PLAN7_DESC) write_bin_string(fp, hmm->desc);
   fwrite((char *) &(hmm->M),        sizeof(int),  1,   fp);
   fwrite((char *) &(Alphabet_type), sizeof(int),  1,   fp);
-  if (hmm->flags & PLAN7_RF)   fwrite((char *) hmm->rf, sizeof(char), hmm->M+1, fp);
-  if (hmm->flags & PLAN7_CS)   fwrite((char *) hmm->cs, sizeof(char), hmm->M+1, fp);
+  if (hmm->flags & PLAN7_RF)   fwrite((char *) hmm->rf,  sizeof(char), hmm->M+1, fp);
+  if (hmm->flags & PLAN7_CS)   fwrite((char *) hmm->cs,  sizeof(char), hmm->M+1, fp);
+  if (hmm->flags & PLAN7_MAP)  fwrite((char *) hmm->map, sizeof(int), hmm->M+1, fp);
   write_bin_string(fp, hmm->comlog);
   fwrite((char *) &(hmm->nseq),     sizeof(int),  1,   fp);
   write_bin_string(fp, hmm->ctime);
+  fwrite((char *) &(hmm->checksum), sizeof(int),  1,   fp);
+  if (hmm->flags & PLAN7_GA) {
+    fwrite((char *) &(hmm->ga1), sizeof(float), 1, fp);
+    fwrite((char *) &(hmm->ga2), sizeof(float), 1, fp);
+  }
+  if (hmm->flags & PLAN7_TC) {
+    fwrite((char *) &(hmm->tc1), sizeof(float), 1, fp);
+    fwrite((char *) &(hmm->tc2), sizeof(float), 1, fp);
+  }
+  if (hmm->flags & PLAN7_NC) {
+    fwrite((char *) &(hmm->nc1), sizeof(float), 1, fp);
+    fwrite((char *) &(hmm->nc2), sizeof(float), 1, fp);
+  }
 
   /* Specials */
   for (k = 0; k < 4; k++)
@@ -496,15 +575,18 @@ read_asc20hmm(HMMFILE *hmmfp, struct plan7_s **ret_hmm)
   M = -1;
   while (fgets(buffer, 512, hmmfp->f) != NULL) {
     if      (strncmp(buffer, "NAME ", 5) == 0) Plan7SetName(hmm, buffer+6);
+    else if (strncmp(buffer, "ACC  ", 5) == 0) Plan7SetAccession(hmm, buffer+6);
     else if (strncmp(buffer, "DESC ", 5) == 0) Plan7SetDescription(hmm, buffer+6);
     else if (strncmp(buffer, "LENG ", 5) == 0) M = atoi(buffer+6);
     else if (strncmp(buffer, "NSEQ ", 5) == 0) hmm->nseq = atoi(buffer+6);
     else if (strncmp(buffer, "ALPH ", 5) == 0) 
       {				/* Alphabet type */
-	s2upper(buffer+6);
-	if      (strncmp(buffer+6, "AMINO",   5) == 0) SetAlphabet(hmmAMINO);
-	else if (strncmp(buffer+6, "NUCLEIC", 7) == 0) SetAlphabet(hmmNUCLEIC);
-	else goto FAILURE;
+	if (Alphabet_type == hmmNOTSETYET) {
+	  s2upper(buffer+6);
+	  if      (strncmp(buffer+6, "AMINO",   5) == 0) SetAlphabet(hmmAMINO);
+	  else if (strncmp(buffer+6, "NUCLEIC", 7) == 0) SetAlphabet(hmmNUCLEIC);
+	  else goto FAILURE;
+	}
       }
     else if (strncmp(buffer, "RF   ", 5) == 0) 
       {				/* Reference annotation present? */
@@ -514,6 +596,10 @@ read_asc20hmm(HMMFILE *hmmfp, struct plan7_s **ret_hmm)
       {				/* Consensus annotation present? */
 	if (sre_toupper(*(buffer+6)) == 'Y') hmm->flags |= PLAN7_CS;
       }
+    else if (strncmp(buffer, "MAP  ", 5) == 0) 
+      {				/* Map annotation present? */
+	if (sre_toupper(*(buffer+6)) == 'Y') hmm->flags |= PLAN7_MAP;
+      }
     else if (strncmp(buffer, "COM  ", 5) == 0) 
       {				/* Command line log */
 	StringChop(buffer+6);
@@ -532,6 +618,30 @@ read_asc20hmm(HMMFILE *hmmfp, struct plan7_s **ret_hmm)
 	StringChop(buffer+6);
 	hmm->ctime= Strdup(buffer+6); 
       }
+    else if (strncmp(buffer, "GA   ", 5) == 0)
+      {
+	if ((s = strtok(buffer+6, " \t\n")) == NULL) goto FAILURE;
+	hmm->ga1 = atof(s);
+	if ((s = strtok(NULL, " \t\n")) == NULL) goto FAILURE;
+	hmm->ga2 = atof(s);
+	hmm->flags |= PLAN7_GA;
+      }
+    else if (strncmp(buffer, "TC   ", 5) == 0)
+      {
+	if ((s = strtok(buffer+6, " \t\n")) == NULL) goto FAILURE;
+	hmm->tc1 = atof(s);
+	if ((s = strtok(NULL, " \t\n")) == NULL) goto FAILURE;
+	hmm->tc2 = atof(s);
+	hmm->flags |= PLAN7_TC;
+      }
+    else if (strncmp(buffer, "NC   ", 5) == 0)
+      {
+	if ((s = strtok(buffer+6, " \t\n")) == NULL) goto FAILURE;
+	hmm->nc1 = atof(s);
+	if ((s = strtok(NULL, " \t\n")) == NULL) goto FAILURE;
+	hmm->nc2 = atof(s);
+	hmm->flags |= PLAN7_NC;
+      }
     else if (strncmp(buffer, "XT   ", 5) == 0) 
       {				/* Special transition section */
 	if ((s = strtok(buffer+6, " \t\n")) == NULL) goto FAILURE;
@@ -564,11 +674,12 @@ read_asc20hmm(HMMFILE *hmmfp, struct plan7_s **ret_hmm)
     else if (strncmp(buffer, "EVD  ", 5) == 0) 
       {				/* EVD parameters */
 	hmm->flags |= PLAN7_STATS;
-	if ((s = strtok(buffer+6, " \t\n")) == NULL)    goto FAILURE;
+	if ((s = strtok(buffer+6, " \t\n")) == NULL) goto FAILURE;
 	hmm->mu = atof(s);
-	if ((s = strtok(NULL, " \t\n")) == NULL)   goto FAILURE;
+	if ((s = strtok(NULL, " \t\n")) == NULL) goto FAILURE;
 	hmm->lambda = atof(s);
       }
+    else if (strncmp(buffer, "CKSUM", 5) == 0) hmm->checksum = atoi(buffer+6);
     else if (strncmp(buffer, "HMM  ", 5) == 0) break;
   }
 
@@ -595,13 +706,17 @@ read_asc20hmm(HMMFILE *hmmfp, struct plan7_s **ret_hmm)
 
 				/* main model */
   for (k = 1; k <= hmm->M; k++) {
-                                /* Line 1: k and match emissions */
+                                /* Line 1: k, match emissions, map */
     if (fgets(buffer, 512, hmmfp->f) == NULL)  goto FAILURE;
     if ((s = strtok(buffer, " \t\n")) == NULL) goto FAILURE;
     if (atoi(s) != k)                          goto FAILURE;
     for (x = 0; x < Alphabet_size; x++) {
       if ((s = strtok(NULL, " \t\n")) == NULL) goto FAILURE;
       hmm->mat[k][x] = ascii2prob(s, hmm->null[x]);
+    }
+    if (hmm->flags & PLAN7_MAP) {
+      if ((s = strtok(NULL, " \t\n")) == NULL) goto FAILURE;
+      hmm->map[k] = atoi(s);
     }
 				/* Line 2:  RF and insert emissions */
     if (fgets(buffer, 512, hmmfp->f) == NULL)  goto FAILURE;
@@ -653,15 +768,15 @@ read_bin20hmm(HMMFILE *hmmfp, struct plan7_s **ret_hmm)
    struct plan7_s *hmm;
    int    k,x;
    int    type;
-   unsigned long magic;
+   unsigned int magic;
 
    hmm = NULL;
 
    /* Header section
     */
    if (feof(hmmfp->f))                                      return 0;
-   if (! fread((char *) &magic, sizeof(long), 1, hmmfp->f)) return 0;
-   if (hmmfp->byteswap) byteswap((char *)&magic, sizeof(long));
+   if (! fread((char *) &magic, sizeof(unsigned int), 1, hmmfp->f)) return 0;
+   if (hmmfp->byteswap) byteswap((char *)&magic, sizeof(unsigned int));
    if (magic != v20magic) goto FAILURE;
 				/* allocate HMM shell for header info */
    hmm = AllocPlan7Shell();
@@ -670,6 +785,9 @@ read_bin20hmm(HMMFILE *hmmfp, struct plan7_s **ret_hmm)
    if (hmmfp->byteswap) byteswap((char *)&(hmm->flags), sizeof(int)); 
 				/* name */
    if (! read_bin_string(hmmfp->f, hmmfp->byteswap, &(hmm->name))) goto FAILURE;
+				/* optional accession */
+   if ((hmm->flags & PLAN7_ACC) &&
+       ! read_bin_string(hmmfp->f, hmmfp->byteswap, &(hmm->acc))) goto FAILURE;
 				/* optional description */
    if ((hmm->flags & PLAN7_DESC) &&
        ! read_bin_string(hmmfp->f, hmmfp->byteswap, &(hmm->desc))) goto FAILURE;
@@ -679,7 +797,7 @@ read_bin20hmm(HMMFILE *hmmfp, struct plan7_s **ret_hmm)
 				/* alphabet type */
    if (! fread((char *) &type, sizeof(int), 1, hmmfp->f)) goto FAILURE;
    if (hmmfp->byteswap) byteswap((char *)&type, sizeof(int)); 
-   if (Alphabet_type == 0) SetAlphabet(type);
+   if (Alphabet_type == hmmNOTSETYET) SetAlphabet(type);
 
 				/* now allocate for rest of model */
    AllocPlan7Body(hmm, hmm->M);
@@ -692,17 +810,61 @@ read_bin20hmm(HMMFILE *hmmfp, struct plan7_s **ret_hmm)
    if ((hmm->flags & PLAN7_CS) &&
        !fread((char *) hmm->cs, sizeof(char), hmm->M+1, hmmfp->f)) goto FAILURE;
    hmm->cs[hmm->M+1]  = '\0';
+				/* optional alignment map annotation */
+   if ((hmm->flags & PLAN7_MAP) &&
+       !fread((char *) hmm->map, sizeof(int), hmm->M+1, hmmfp->f)) goto FAILURE;
+   if (hmmfp->byteswap)
+     for (k = 1; k <= hmm->M; k++)
+       byteswap((char*)&(hmm->map[k]), sizeof(int));
 				/* command line log */
    if (!read_bin_string(hmmfp->f, hmmfp->byteswap, &(hmm->comlog)))  goto FAILURE;
 				/* nseq */
    if (!fread((char *) &(hmm->nseq),sizeof(int), 1, hmmfp->f))       goto FAILURE;
+   if (hmmfp->byteswap) byteswap((char *)&(hmm->nseq), sizeof(int)); 
 				/* creation time */
    if (!read_bin_string(hmmfp->f, hmmfp->byteswap, &(hmm->ctime)))   goto FAILURE;
+				/* checksum */
+   if (!fread((char *) &(hmm->checksum),sizeof(int), 1, hmmfp->f))       goto FAILURE;
+   if (hmmfp->byteswap) byteswap((char *)&(hmm->checksum), sizeof(int)); 
      
-  /* specials */
-   for (k = 0; k < 4; k++)
-     if (! fread((char *) hmm->xt[k], sizeof(float), 2, hmmfp->f))    goto FAILURE;
+				/* Pfam gathering thresholds */
+   if (hmm->flags & PLAN7_GA) {
+     if (! fread((char *) &(hmm->ga1), sizeof(float), 1, hmmfp->f)) goto FAILURE;
+     if (! fread((char *) &(hmm->ga2), sizeof(float), 1, hmmfp->f)) goto FAILURE;
+     if (hmmfp->byteswap) {
+       byteswap((char *) &(hmm->ga1), sizeof(float));
+       byteswap((char *) &(hmm->ga2), sizeof(float));
+     }
+   }
+				/* Pfam trusted cutoffs */
+   if (hmm->flags & PLAN7_TC) {
+     if (! fread((char *) &(hmm->tc1), sizeof(float), 1, hmmfp->f)) goto FAILURE;
+     if (! fread((char *) &(hmm->tc2), sizeof(float), 1, hmmfp->f)) goto FAILURE;
+     if (hmmfp->byteswap) {
+       byteswap((char *) &(hmm->tc1), sizeof(float));
+       byteswap((char *) &(hmm->tc2), sizeof(float));
+     }
+   }
+				/* Pfam noise cutoffs */
+   if (hmm->flags & PLAN7_NC) {
+     if (! fread((char *) &(hmm->nc1), sizeof(float), 1, hmmfp->f)) goto FAILURE;
+     if (! fread((char *) &(hmm->nc2), sizeof(float), 1, hmmfp->f)) goto FAILURE;
+     if (hmmfp->byteswap) {
+       byteswap((char *) &(hmm->nc1), sizeof(float));
+       byteswap((char *) &(hmm->nc2), sizeof(float));
+     }
+   }
 
+   /* specials */
+   for (k = 0; k < 4; k++)
+     {
+       if (! fread((char *) hmm->xt[k], sizeof(float), 2, hmmfp->f))    goto FAILURE;
+       if (hmmfp->byteswap) {
+	 for (x = 0; x < 2; x++)
+	   byteswap((char *)&(hmm->xt[k][x]), sizeof(float));
+       }
+     }
+   
    /* null model */
    if (!fread((char *) &(hmm->p1),sizeof(float), 1, hmmfp->f))        goto FAILURE;
    if (!fread((char *)hmm->null,sizeof(float),Alphabet_size,hmmfp->f))goto FAILURE;
@@ -749,8 +911,9 @@ read_bin20hmm(HMMFILE *hmmfp, struct plan7_s **ret_hmm)
 	    byteswap((char *)&(hmm->ins[k][x]), sizeof(float));
 	byteswap((char *)&(hmm->begin[k]),  sizeof(float));
 	byteswap((char *)&(hmm->end[k]),    sizeof(float));
-	for (x = 0; x < 7; x++) 
-	  byteswap((char *)&(hmm->t[k][x]), sizeof(float));
+	if (k < hmm->M)
+	  for (x = 0; x < 7; x++) 
+	    byteswap((char *)&(hmm->t[k][x]), sizeof(float));
       }
   }
 
@@ -803,9 +966,11 @@ read_asc19hmm(HMMFILE *hmmfp, struct plan7_s **ret_hmm)
   if ((s = Getword(fp, sqdARG_INT))    == NULL) goto FAILURE;                        /* ignore alphabet size */
   if ((s = Getword(fp, sqdARG_STRING)) == NULL) goto FAILURE;  Plan7SetName(hmm, s); /* name */
   if ((s = Getword(fp, sqdARG_STRING)) == NULL) goto FAILURE;  s2upper(s);           /* alphabet type */ 
-  if      (strcmp(s, "AMINO") == 0)   SetAlphabet(hmmAMINO);
-  else if (strcmp(s, "NUCLEIC") == 0) SetAlphabet(hmmNUCLEIC);
-  else goto FAILURE;
+  if (Alphabet_type == hmmNOTSETYET) {
+    if      (strcmp(s, "AMINO") == 0)   SetAlphabet(hmmAMINO);
+    else if (strcmp(s, "NUCLEIC") == 0) SetAlphabet(hmmNUCLEIC);
+    else goto FAILURE;
+  }
 				/* read alphabet, make sure it's Plan7-compatible... */
   if ((s = Getword(fp, sqdARG_STRING)) == NULL) goto FAILURE;
   if (strncmp(s, Alphabet, Alphabet_size) != 0) goto FAILURE;
