@@ -1,11 +1,11 @@
 /************************************************************
- * HMMER - Biological sequence analysis with profile-HMMs
+ * HMMER - Biological sequence analysis with profile HMMs
  * Copyright (C) 1992-1998 Washington University School of Medicine
- *
- *   This source code is distributed under the terms of the 
- *   GNU General Public License. See the files COPYING and 
- *   GNULICENSE for details.
- *    
+ * All Rights Reserved
+ * 
+ *     This source code is distributed under the terms of the
+ *     GNU General Public License. See the files COPYING and LICENSE
+ *     for details.
  ************************************************************/
 
 /* structs.h
@@ -13,7 +13,7 @@
  * Data structures used in HMMER.
  * Also, a few miscellaneous macros and global variable declarations.
  * 
- * RCS $Id: structs.h,v 1.10 1998/05/27 02:44:50 eddy Exp $
+ * RCS $Id: structs.h,v 1.17 1998/12/09 17:42:36 eddy Exp $
  */
 
 #ifndef STRUCTSH_INCLUDED
@@ -35,17 +35,6 @@
  */
 #define SYMIDX(x)   (strchr(Alphabet, (x)) - Alphabet)
 
-/* Worry() is a conditional warning macro, used for debugging;
- * it inserts file name and line number information and
- * calls VerboseWorry(). See debug.c. 
- * 
- * PANIC is called for failures of Std C/POSIX functions,
- * instead of my own functions. It calls perror() and exits
- * abnormally.
- */
-#define Worry(x,s)  VerboseWorry(x, __FILE__, __LINE__, s)
-#define PANIC       (Panic(__FILE__, __LINE__))
-
 /* The symbol alphabet.
  * Must deal with IUPAC degeneracies. Nondegenerate symbols 
  * come first in Alphabet[], followed by degenerate symbols.
@@ -67,7 +56,7 @@ extern int   Alphabet_iupac;    /* total size of alphabet + IUPAC degen. */
 extern char  Degenerate[MAXCODE][MAXABET];
 extern int   DegenCount[MAXCODE];
 #define hmmNOTSETYET 0
-#define hmmNUCLEIC   2		/* compatibility with squid's kDNA   */
+#define hmmNUCLEIC   2		/* compatibility with squid's kRNA   */
 #define hmmAMINO     3		/* compatibility with squid's kAmino */
 
 /**********************************************************************
@@ -109,17 +98,32 @@ struct plan7_s {
    * Other fields are optional; whether they are present is
    * flagged in the stateflags bit array.
    * 
-   * desc is only valid if PLAN7_DESC is set.
-   *   rf is only valid if PLAN7_RF is set.
-   *   cs is only valid if PLAN7_CS is set.
+   * desc is only valid if PLAN7_DESC is set in flags.
+   *  acc is only valid if PLAN7_ACC is set in flags.
+   *   rf is only valid if PLAN7_RF is set in flags.
+   *   cs is only valid if PLAN7_CS is set in flags.
+   *  map is only valid if PLAN7_MAP is set in flags.
    */
   char  *name;                  /* name of the model                    +*/
+  char  *acc;			/* accession number of model (Pfam)     +*/
   char  *desc;                  /* brief description of model           +*/ 
   char  *rf;                    /* reference line from alignment 0..M   +*/
   char  *cs;                    /* consensus structure line      0..M   +*/ 
   char  *comlog;		/* command line(s) that built model     +*/
   int    nseq;			/* number of training sequences         +*/
   char  *ctime;			/* creation date                        +*/
+  int   *map;			/* map of alignment cols onto model 1..M+*/
+  int    checksum;              /* checksum of training sequences       +*/
+
+  /* Pfam-specific score cutoffs.
+   * 
+   * ga1, ga2 are valid if PLAN7_GA is set in flags.
+   * tc1, tc2 are valid if PLAN7_TC is set in flags.
+   * nc1, nc2 are valid if PLAN7_NC is set in flags.
+   */
+  float  ga1, ga2;		/* per-seq/per-domain gathering thresholds (bits) +*/
+  float  tc1, tc2;		/* per-seq/per-domain trusted cutoff (bits)       +*/
+  float  nc1, nc2;		/* per-seq/per-domain noise cutoff (bits)         +*/
 
   /* The main model in probability form: data-dependent probabilities.
    * This is the core Krogh/Haussler model.
@@ -136,6 +140,10 @@ struct plan7_s {
 
   /* The unique states of Plan 7 in probability form.
    * These are the algorithm-dependent, data-independent probabilities.
+   * Some parts of the code may briefly use a trick of copying tbd1
+   * into begin[0]; this makes it easy to call FChoose() or FNorm()
+   * on the resulting vector. However, in general begin[0] is not
+   * a valid number.
    */
   float  xt[4][2];              /* N,E,C,J extra states: 2 transitions      +*/
   float *begin;                 /* 1..M B->M state transitions              +*/
@@ -183,9 +191,8 @@ struct plan7_s {
    */
   float  mu;			/* EVD mu       +*/
   float  lambda;		/* EVD lambda   +*/
-  float  wonka;			/* EVD fit display fudge factor +*/
 
-  int flags;                    /* bit flags indicating state of HMM    +*/
+  int flags;                    /* bit flags indicating state of HMM, valid data +*/
 };
 
 /* Flags for plan7->flags.
@@ -205,6 +212,11 @@ struct plan7_s {
 #define PLAN7_HASPROB (1<<5)    /* raised if model has probabilities        */
 #define PLAN7_HASDNA  (1<<6)	/* raised if protein HMM->DNA seq params set*/
 #define PLAN7_STATS   (1<<7)	/* raised if EVD parameters are available   */
+#define PLAN7_MAP     (1<<8)	/* raised if alignment map is available     */
+#define PLAN7_ACC     (1<<9)	/* raised if accession number is available  */
+#define PLAN7_GA      (1<<10)	/* raised if gathering thresholds available */
+#define PLAN7_TC      (1<<11)	/* raised if trusted cutoffs available      */
+#define PLAN7_NC      (1<<12)	/* raised if noise cutoffs available        */
 
 /* Indices for special state types, I: used for dynamic programming xmx[][]
  * mnemonic: eXtra Matrix for B state = XMB
@@ -250,15 +262,29 @@ struct dpmatrix_s {
   int **dmx;			/* delete scores [0.1..N][0.1..M-1.M] */
 };
 
+/* Declaration of Plan7 shadow matrix structure.
+ * In general, allowed values are STM, STI, etc.
+ * However, E state has M possible sources, from 1..M match states;
+ * hence the esrc array.
+ */
+struct dpshadow_s {
+  char **xtb;			/* special state traces [0.1..N][BECJN]     */
+  char **mtb;			/* match state traces [0.1..N][0.1..M]      */
+  char **itb;			/* insert state traces [0.1..N][0.1..M-1.M] */
+  char **dtb;			/* delete state traces [0.1..N][0.1..M-1.M] */
+  int   *esrc;                  /* E trace is special; must store a M state number 1..M */
+};
+
 /* Structure: HMMFILE
  * 
  * Purpose:   An open HMM file or HMM library. See hmmio.c.
  */
 struct hmmfile_s {
-  FILE *f;			/* pointer to file opened for reading     */
+  FILE    *f;			/* pointer to file opened for reading       */
+  GSIFILE *gsi;			/* pointer to open GSI index, or NULL       */
   int (*parser)(struct hmmfile_s *, struct plan7_s **);  /* parsing function*/
-  int   is_binary;		/* TRUE if format is a binary one         */
-  int   byteswap;               /* TRUE if binary and byteswapped         */
+  int   is_binary;		/* TRUE if format is a binary one           */
+  int   byteswap;               /* TRUE if binary and byteswapped           */
 };
 typedef struct hmmfile_s HMMFILE; 
 
@@ -266,7 +292,17 @@ typedef struct hmmfile_s HMMFILE;
 /* Plan 7 model state types
  * used in traceback structure
  */
-enum p7stype { STM, STD, STI, STS, STN, STB, STE, STC, STT, STJ, STBOGUS };
+#define STBOGUS 0
+#define STM     1
+#define STD     2
+#define STI     3
+#define STS     4
+#define STN     5
+#define STB     6
+#define STE     7
+#define STC     8
+#define STT     9
+#define STJ     10     
 
 /* Structure: p7trace_s
  * 
@@ -276,7 +312,7 @@ enum p7stype { STM, STD, STI, STS, STN, STB, STE, STC, STT, STJ, STBOGUS };
  */
 struct p7trace_s {
   int   tlen;                   /* length of traceback                      */
-  enum p7stype *statetype;      /* state type used for alignment            */
+  char *statetype;              /* state type used for alignment            */
   int  *nodeidx;                /* index of aligned node, 1..M (if M,D,I)   */
   int  *pos;                    /* position in dsq, 1..L or -1              */ 
 };
@@ -404,10 +440,32 @@ struct tophit_s {
   int            lump;       	/* allocation lumpsize                      */
 };
 
+/**********************************************************
+ * PVM parallelization
+ **********************************************************/
+#ifdef  HMMER_PVM
+
+/* Message tags 
+ */
+#define HMMPVM_INIT         0	/* an initialization packet to all slaves */
+#define HMMPVM_WORK         1	/* a work packet sent to a slave */
+#define HMMPVM_RESULTS      2	/* a results packet sent back to master */
+#define HMMPVM_TASK_TROUBLE 3	/* a notification of bad things in a slave task */
+#define HMMPVM_HOST_TROUBLE 4	/* a notification of bad things in a PVM host */
+
+/* error codes
+ */
+#define HMMPVM_OK         0
+#define HMMPVM_NO_HMMFILE 1
+#define HMMPVM_NO_INDEX   2	
+#define HMMPVM_BAD_INIT   3	/* failed to initialize a slave somehow */
+
+#endif
+
 
 /**********************************************************
  * Plan 9: obsolete HMMER1.x code. We still need these structures
- * for reading old HMM files (e.g. reverse compatibility)
+ * for reading old HMM files (e.g. backwards compatibility)
  **********************************************************/
 
 /* We define a "basic" state, which covers the basic match, insert, and
